#if (defined(__IBMCPP__) || defined(__EMX__))
   #define INCL_DOSPROCESS
   #include <os2.h>
#endif
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <string.h>
#include <stdlib.h>

#include "TWRemaps.hpp"
#include "TWTypes.hpp"
#include "Globals.hpp"
#include "TWSector.hpp"
#include "TWWarp.hpp"
#include "TWBubble.hpp"
#include "AbEnds.hpp"
#include "TWBFind.h"

int main(int argc, char* argv[], char* envp[])
{
   char sectFileName[256], outFileName[256], argtype[3], argdata[256],
        argtemp[256];
   USHORT numBubbles, i, mxetemp;
   ifstream sectFile;
   ofstream outFile;
   TWBubbleSet bubbleSet, keepSet;

   strcpy(sectFileName, "");
   strcpy(outFileName, "bubbles.lst");

   for (i=1; i<argc; i++)
      {
      memset(argtype, 0, sizeof(argtype));
      memset(argdata, 0, sizeof(argdata));
      memset(argtemp, 0, sizeof(argtemp));
      strncpy(argtype, argv[i], 2);
      strncpy(argtemp, strrev(argv[i]), (strlen(argv[i])-2));
      strcpy(argdata, strrev(argtemp));
      if (!strcmpi(argtype, "/t") || !strcmpi(argtype, "-t"))
         {
         Globals::tunnelEntrance=true;
         }
      else if (!strcmpi(argtype, "/m") || !strcmpi(argtype, "-m"))
         {
         Globals::minSectors=(USHORT)atoi(argdata);
         if (Globals::minSectors==0)
            BadUsage(argdata, BAD_MIN_SECTORS);
         }
      else if (!strcmpi(argtype, "/x") || !strcmpi(argtype, "-x"))
         {
         Globals::maxSectors=(USHORT)atoi(argdata);
         if (Globals::maxSectors==0)
            BadUsage(argdata, BAD_MAX_SECTORS);
         }
      else if (!strcmpi(argtype, "/z") || !strcmpi(argtype, "-z"))
         {
         Globals::entMultiplier=(float)atof(argdata);
         if (Globals::entMultiplier<2.0 ||
             Globals::entMultiplier>1000.0)
            BadUsage(argdata, BAD_ENT_MULTIPLIER);
         }
      else if (!strcmpi(argtype, "/s") || !strcmpi(argtype, "-s"))
         {
         memset(sectFileName, 0, sizeof(sectFileName));
         strcpy(sectFileName, argdata);
         }
      else if (!strcmpi(argtype, "/o") || !strcmpi(argtype, "-o"))
         {
         memset(outFileName, 0, sizeof(outFileName));
         strcpy(outFileName, argdata);
         }
      else if (!strcmpi(argtype, "/n") || !strcmpi(argtype, "-n"))
         {
         Globals::totalSectors=(USHORT)atoi(argdata);
         if (Globals::totalSectors==0)
            BadUsage(argdata, BAD_TOTAL_SECTORS);
         }
      else if (!strcmpi(argtype, "/e") || !strcmpi(argtype, "-e"))
         {
         mxetemp=(USHORT)atoi(argdata);
         if (mxetemp==0 || mxetemp>50)
            BadUsage(argdata, BAD_MAX_ENTRANCES);
         else
            Globals::maxEntrances=(UCHAR)mxetemp;
         }
      #if (defined(__OS2__) || defined(__EMX__)) // priority options
      else if (!strcmpi(argtype, "/p") || !strcmpi(argtype, "-p"))
         {
         Globals::pclass=(ULONG)atol(argdata);
         if (Globals::pclass==0 || Globals::pclass>4)
            BadUsage(argdata, BAD_PRIORITY_CLASS);
         }
      else if (!strcmpi(argtype, "/d") || !strcmpi(argtype, "-d"))
         {
         if (!strcmpi(argdata, "0"))
            {
            Globals::pdelta=0;
            }
         else
            {
            Globals::pdelta=(LONG)atol(argdata);
            if (Globals::pdelta==0 || Globals::pdelta<-31 || Globals::pdelta>31)
               BadUsage(argdata, BAD_PRIORITY_DELTA);
            }
         }
      #endif
      else if (!strcmpi(argtype, "/h") || !strcmpi(argtype, "-h") ||
               !strcmpi(argtype, "/?") || !strcmpi(argtype, "-?"))
         {
         BadUsage("", USAGE_HELP);
         }
      else
         BadUsage(argtype, BAD_PARAMETER);
      }

   if (Globals::maxEntrances>Globals::maxSectors ||
       Globals::maxEntrances>(Globals::maxSectors-Globals::minSectors))
      {
      memset(argtemp, 0, sizeof(argtemp));
      BadUsage(_itoa(Globals::maxEntrances, argtemp, 10), BAD_MAX_ENTRANCES);
      }
   if (Globals::minSectors>Globals::maxSectors ||
       Globals::minSectors>Globals::totalSectors ||
       Globals::minSectors<2)
      {
      memset(argtemp, 0, sizeof(argtemp));
      BadUsage(_itoa(Globals::minSectors, argtemp, 10), BAD_MIN_SECTORS);
      }
   if (Globals::maxSectors>Globals::totalSectors || Globals::maxSectors<2)
      {
      memset(argtemp, 0, sizeof(argtemp));
      BadUsage(_itoa(Globals::maxSectors, argtemp, 10), BAD_MAX_SECTORS);
      }
   if (!strcmp(sectFileName,""))
      BadUsage("", USAGE_HELP);

   sectFile.open(sectFileName, ios::nocreate | ios::in);
   if (sectFile.fail()!=0)
      BadUsage(sectFileName, BAD_SECTOR_FILE);
   outFile.open(outFileName, ios::trunc | ios::out);
   if (outFile.fail()!=0)
      BadUsage(outFileName, BAD_REPORT_FILE);

   if (!Globals::sectors.setNumberOfSectors(Globals::totalSectors))
      OutOfMemory();
   cout << endl;
   cout << "Reading warp data for sector:      ";
   Globals::sectors.readWarpData(sectFile);
   sectFile.close();

   cout << endl << endl;
   cout << "Setting one-way warp status for sector:      ";
   Globals::sectors.setOneWayStatus();

   #if (defined(__OS2__) || defined(__EMX__))
   DosSetPriority(2, Globals::pclass, Globals::pdelta, 0);
   #endif

   cout << endl << endl;
   cout << "Prospecting potential bubbles with entrance :      ";
   Globals::sectors.findBubbles(bubbleSet);

   if (bubbleSet.numberOfBubbles()>1)
      {
      cout << endl << endl;
      cout << "Trimming " << bubbleSet.numberOfBubbles() << " potential bubbles..." << flush;
      }
   bubbleSet.trimBubblesTo(keepSet);

   cout << endl << endl;
   numBubbles=keepSet.numberOfBubbles();
   cout << "Number of bubbles found: " << numBubbles << endl << flush;

   for (i=0; i<numBubbles; i++)
      {
      keepSet[i].writeTo(outFile);
      }

   outFile.close();

   #ifdef DEBUG_ENABLED
   Globals::debug.close();
   #endif
   return 0;
}
