#include "TWBubble.hpp"
#include "TWSector.hpp"
#include "TWWarp.hpp"
#include "AbEnds.hpp"
#include "Globals.hpp"
#include <fstream.h>
#include <iomanip.h>
#include <memory.h>

TWBubble::TWBubble()
{
   sectors=0;
   numSectors=0;
   entrances=0;
   numBD=0;
   numEnts=0;
   back=0;
}

TWBubble::TWBubble(USHORT entrance)
{
   sectors = new USHORT[1];
   if (sectors==NULL)
      OutOfMemory();
   sectors[0]=entrance;
   numSectors=1;
   entrances = new USHORT[1];
   if (entrances==NULL)
      OutOfMemory();
   entrances[0]=entrance;
   numEnts=1;
   numBD=0;
   back=0;
}

TWBubble::TWBubble(TWBubble& copy)
{
   numSectors=copy.numSectors;
   numEnts=copy.numEnts;
   numBD=copy.numBD;
   if (numSectors>0)
      {
      sectors = new USHORT[numSectors];
      if (sectors==NULL)
         OutOfMemory();
      memcpy(sectors, copy.sectors, sizeof(USHORT)*numSectors);
      }
   else sectors=0;
   if (numEnts>0)
      {
      entrances = new USHORT[numEnts];
      if (entrances==NULL)
         OutOfMemory();
      memcpy(entrances, copy.entrances, sizeof(USHORT)*numEnts);
      }
   else entrances=0;
   if (numBD>0)
      {
      back = new USHORT[numBD];
      if (back==NULL)
         OutOfMemory();
      memcpy(back, copy.back, sizeof(USHORT)*numBD);
      }
   else back=0;
}

TWBubble& TWBubble::operator=(const TWBubble& assign)
{
   if (this==&assign)
      return *this;
   if (numSectors>0)
      delete[] sectors;
   if (numBD>0)
      delete[] back;
   if (numEnts>0)
      delete[] entrances;
   numBD=assign.numBD;
   numSectors=assign.numSectors;
   numEnts=assign.numEnts;
   if (numSectors>0)
      {
      sectors = new USHORT[numSectors];
      if (sectors==NULL)
         OutOfMemory();
      memcpy(sectors, assign.sectors, sizeof(USHORT)*numSectors);
      }
   else
      sectors=0;
   if (numBD>0)
      {
      back = new USHORT[numBD];
      if (back==NULL)
         OutOfMemory();
      memcpy(back, assign.back, sizeof(USHORT)*numBD);
      }
   else
      back=0;
   if (numEnts>0)
      {
      entrances = new USHORT[numEnts];
      if (entrances==NULL)
         OutOfMemory();
      memcpy(entrances, assign.entrances, sizeof(USHORT)*numEnts);
      }
   else
      entrances=0;
   return *this;
}

Bool TWBubble::operator==(const TWBubble& comp) const
{
   if (numSectors!=comp.numSectors || numBD!=comp.numBD || numEnts!=comp.numEnts)
      return false;
   for (USHORT i=0; i<numSectors; i++)
      {
      if (sectors[i]!=comp.sectors[i])
         {
         return false;
         }
      }
   return true;
}

Bool TWBubble::containsBubble(const TWBubble& sBub) const
{
   if (numSectors<sBub.numSectors)
      return false;
   for (USHORT i=0; i<sBub.numSectors; i++)
      {
      if (!hasSector(sBub.sectors[i]))
         return false;
      }
   if (numEnts>sBub.numEnts)
      return false;
   else
      return true;
}

TWBubble::~TWBubble()
{
   if (numSectors>0)
      delete[] sectors;
   if (numBD>0)
      delete[] back;
   if (numEnts>0)
      delete[] entrances;
}

USHORT TWBubble::numberOfSectors(void) const
{
   return numSectors;
}

USHORT TWBubble::entranceAtPosition(USHORT index) const
{
   return entrances[index];
}

USHORT TWBubble::numberOfEntrances(void) const
{
   return numEnts;
}

Bool TWBubble::addSector(USHORT newSect)
{
   if (numSectors==0)
      {
      sectors = new USHORT[1];
      if (sectors==NULL)
         return false;
      sectors[0]=newSect;
      numSectors++;
      }
   else
      {
      USHORT *tempArray = new USHORT[numSectors+1];
      if (tempArray==NULL)
         return false;
      memcpy(tempArray, sectors, sizeof(USHORT)*numSectors);
      tempArray[numSectors]=newSect;
      numSectors++;
      delete[] sectors;
      sectors=tempArray;
      }
   return true;
}

Bool TWBubble::addEntrance(USHORT newEnt)
{
   if (newEnt==0)
      return true;
   if (numEnts==0)
      {
      entrances = new USHORT[1];
      if (entrances==NULL)
         return false;
      entrances[0]=newEnt;
      numEnts++;
      }
   else
      {
      USHORT *tempArray = new USHORT[numEnts+1];
      if (tempArray==NULL)
         return false;
      memcpy(tempArray, entrances, sizeof(USHORT)*numEnts);
      tempArray[numEnts]=newEnt;
      numEnts++;
      delete[] entrances;
      entrances=tempArray;
      }
   return true;
}


Bool TWBubble::hasSector(USHORT sectNum) const
{
   if (numSectors==0)
      return false;
   for (USHORT i=0; i<numSectors; i++)
      {
      if (sectors[i]==sectNum)
         return true;
      }
   return false;
}

Bool TWBubble::hasEntrance(USHORT sectNum) const
{
   if (numEnts==0)
      return false;
   for (USHORT i=0; i<numEnts; i++)
      {
      if (entrances[i]==sectNum)
         return true;
      }
   return false;
}

Bool TWBubble::addBackDoor(USHORT sectNum)
{
   if (numBD==0)
      {
      back = new USHORT[1];
      if (back==NULL)
         return false;
      back[0]=sectNum;
      numBD++;
      }
   else
      {
      USHORT *tempArray = new USHORT[numBD+1];
      if (tempArray==NULL)
         return false;
      memcpy(tempArray, back, sizeof(USHORT)*numBD);
      tempArray[numBD]=sectNum;
      numBD++;
      delete[] back;
      back=tempArray;
      }
   return true;
}

Bool TWBubble::isBackDoor(USHORT sectNum) const
{
   if (numBD==0)
      return false;
   for (USHORT i=0; i<numBD; i++)
      {
      if (back[i]==sectNum)
         return true;
      }
   return false;
}

USHORT TWBubble::sectorAtPosition(USHORT index) const
{
   return sectors[index];
}

USHORT TWBubble::operator[](USHORT index) const
{
   return sectors[index];
}

void TWBubble::writeTo(ofstream& outFile) const
{
   USHORT sectNum, i, j;
   UCHAR warps, sectsOnLine=0;
   TWSector tsect;
   TWWarp currWarp;

   outFile << (numSectors-numEnts) << "-sector bubble with ";
   if (numEnts==1)
      {
      outFile << "entrance at " << sectors[0] << "." << endl << endl;
      }
   else
      {
      outFile << numEnts << " entrances at:" << endl;
      for (i=0; i<numEnts; i++)
         {
         sectsOnLine++;
         outFile << setw(6) << entrances[i];
         if (sectsOnLine==12 || i==(numEnts-1))
            {
            sectsOnLine=0;
            outFile << endl;
            }
         }
      outFile << endl;
      }
   outFile << "Sectors in bubble:" << endl;;
   sectsOnLine=0;
   for (i=1; i<numSectors; i++)
      {
      sectNum=(*this)[i];
      if (hasEntrance(sectNum))
         {
         if (i==(numSectors-1) && sectsOnLine!=0)
            outFile << endl;
         continue;
         }
      else if (isBackDoor(sectNum))
         {
         outFile << "(" << sectNum << ") ";
         }
      else
         outFile << " " << sectNum << "  ";
      sectsOnLine++;
      if (sectsOnLine==10 || i==(numSectors-1))
         {
         sectsOnLine=0;
         outFile << endl;
         }
      }
   outFile << endl;
   for (i=0; i<numSectors; i++)
      {
      sectNum=(*this)[i];
      tsect=Globals::sectors[sectNum-1];
      outFile << "Sector " << setw(5) << sectNum << "  -  ";
      warps=tsect.numberOfInWarps();
      for (j=0; j<warps; j++)
         {
         currWarp=tsect.inWarpAtPosition(j);
         if (!currWarp.isOneWay())
            continue;
         outFile << "<" << setw(5) << currWarp.sector() << "> ";
         }
      warps=tsect.numberOfOutWarps();
      for (j=0; j<warps; j++)
         {
         currWarp=tsect.outWarpAtPosition(j);
         if (currWarp.isOneWay())
            {
            if (!Globals::sectors[currWarp.sector()-1].isExplored())
               outFile << "[";
            else
               outFile << "(";
            }
         else if (hasEntrance(currWarp.sector()))
            {
            outFile << "{";
            }
         else
            outFile << " ";
         outFile << setw(5) << currWarp.sector();
         if (currWarp.isOneWay())
            {
            if (!Globals::sectors[currWarp.sector()-1].isExplored())
               outFile << "] ";
            else
               outFile << ") ";
            }
         else if (hasEntrance(currWarp.sector()))
            {
            outFile << "}";
            }
         else
            outFile << "  ";
         }
      outFile << endl;
      }
   outFile << endl
           << "----------------------------------------"
           << "----------------------------------------"
           << endl << endl;
}

TWBubbleSet::TWBubbleSet()
{
   numBubbles=0;
   bubbles=0;
}

TWBubbleSet::~TWBubbleSet()
{
   if (numBubbles>0)
      delete[] bubbles;
}

TWBubbleSet& TWBubbleSet::operator=(const TWBubbleSet& assign)
{
   if (this==&assign)
      return *this;
   if (numBubbles>0)
      delete[] bubbles;
   numBubbles=assign.numBubbles;
   if (numBubbles>0)
      {
      bubbles = new TWBubble[numBubbles];
      if (bubbles==NULL)
         OutOfMemory();
      for (USHORT i=0; i<numBubbles; i++)
         {
         bubbles[i]=assign.bubbles[i];
         }
      }
   else
      bubbles=0;
   return *this;
}


Bool TWBubbleSet::addBubble(const TWBubble& newBubble)
{
   if (numBubbles==0)
      {
      bubbles = new TWBubble[1];
      if (bubbles==NULL)
         return false;
      numBubbles++;
      bubbles[0]=newBubble;
      }
   else
      {
      TWBubble *tempArray = new TWBubble[numBubbles+1];
      if (tempArray==NULL)
         return false;
      for (USHORT i=0; i<numBubbles; i++)
         {
         tempArray[i]=bubbles[i];
         }
      tempArray[numBubbles]=newBubble;
      numBubbles++;
      delete[] bubbles;
      bubbles=tempArray;
      }
   return true;
}

USHORT TWBubbleSet::numberOfBubbles(void)
{
   return numBubbles;
}

TWBubble& TWBubbleSet::bubbleAtPosition(USHORT index)
{
   return bubbles[index];
}

TWBubble& TWBubbleSet::operator[](USHORT index)
{
   return bubbles[index];
}

void TWBubbleSet::trimBubblesTo(TWBubbleSet& keepSet)
{
   USHORT numBubbles, i, j;
   Bool biggestBubble;

   numBubbles=numberOfBubbles();
   for (i=0; i<numBubbles; i++)
      {
      biggestBubble=true;
      for (j=0; j<numBubbles; j++)
         {
         if (i==j)
            continue;
         if (bubbles[i]==bubbles[j])
            {
            if (j<i)
               {
               continue;
               }
            else
               {
               biggestBubble=false;
               break;
               }
            }
         else if (bubbles[j].containsBubble(bubbles[i]))
            {
            biggestBubble=false;
            break;
            }
         }
      if (biggestBubble)
         {
         keepSet.addBubble(bubbles[i]);
         }
      }
}
