#if (defined(__OS2__) || defined(__EMX__))
   #define INCL_DOSPROCESS
   #include <os2.h>
#endif
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>

#include "Globals.hpp"
#include "TWSector.hpp"
#include "TWBubble.hpp"
#include "AbEnds.hpp"
#include "TWBFind.h"

#if (!defined(__OS2__))  // Add your compiler's define if it has _getch and _kbhit
   #define _getch() getch()
   #define _kbhit() kbhit()
#endif

TWSector::TWSector()
{
   sdata = new TWSectorData;
   if (sdata==NULL)
      OutOfMemory();
   sdata->numInWarps=0;
   sdata->numOutWarps=0;
   sdata->inWarps=0;
   sdata->outWarps=0;
   sdata->explored=false;
}

TWSector::~TWSector()
{
   if (sdata->numInWarps>0)
      delete[] sdata->inWarps;
   if (sdata->numOutWarps>0)
      delete[] sdata->outWarps;
   delete sdata;
}

TWSector::TWSector(TWSector& copy)
{
   sdata = new TWSectorData;
   if (sdata==NULL)
      OutOfMemory();
   sdata->numInWarps=copy.sdata->numInWarps;
   if (sdata->numInWarps>0)
      {
      sdata->inWarps = new TWWarp[sdata->numInWarps];
      if (sdata->inWarps==NULL)
         OutOfMemory();
      memcpy(sdata->inWarps, copy.sdata->inWarps, sizeof(TWWarp)*sdata->numInWarps);
      }
   else
      sdata->inWarps=0;
   sdata->numOutWarps=copy.sdata->numOutWarps;
   if (sdata->numOutWarps>0)
      {
      sdata->outWarps = new TWWarp[sdata->numOutWarps];
      if (sdata->outWarps==NULL)
         OutOfMemory();
      memcpy(sdata->outWarps, copy.sdata->outWarps, sizeof(TWWarp)*sdata->numOutWarps);
      }
   else
      sdata->outWarps=0;
   sdata->explored=copy.sdata->explored;
}

TWSector& TWSector::operator=(const TWSector& assign)
{
   if (this==&assign)
      return *this;
   if (sdata->numInWarps>0)
      delete[] sdata->inWarps;
   if (sdata->numOutWarps>0)
      delete[] sdata->outWarps;
   sdata->numInWarps=assign.sdata->numInWarps;
   sdata->numOutWarps=assign.sdata->numOutWarps;
   if (sdata->numInWarps>0)
      {
      sdata->inWarps = new TWWarp[sdata->numInWarps];
      if (sdata->inWarps==NULL)
         OutOfMemory();
      memcpy(sdata->inWarps, assign.sdata->inWarps, sizeof(TWWarp)*sdata->numInWarps);
      }
   else
      sdata->inWarps=0;
   if (sdata->numOutWarps>0)
      {
      sdata->outWarps = new TWWarp[sdata->numOutWarps];
      if (sdata->outWarps==NULL)
         OutOfMemory();
      memcpy(sdata->outWarps, assign.sdata->outWarps, sizeof(TWWarp)*sdata->numOutWarps);
      }
   else
      sdata->inWarps=0;
   sdata->explored=assign.sdata->explored;
   return *this;
}

UCHAR TWSector::numberOfInWarps(void) const
{
   return sdata->numInWarps;
}

UCHAR TWSector::numberOfOutWarps(void) const
{
   return sdata->numOutWarps;
}

TWWarp& TWSector::inWarpAtPosition(UCHAR index)
{
   return sdata->inWarps[index];
}

TWWarp& TWSector::outWarpAtPosition(UCHAR index)
{
   return sdata->outWarps[index];
}

Bool TWSector::addInWarp(const TWWarp& newSect)
{
   if (sdata->numInWarps==0)
      {
      sdata->inWarps = new TWWarp[1];
      if (sdata->inWarps==NULL)
         return false;
      sdata->numInWarps++;
      sdata->inWarps[0]=newSect;
      }
   else
      {
      TWWarp *tempArray = new TWWarp[sdata->numInWarps+1];
      if (tempArray==NULL)
         return false;
      memcpy(tempArray, sdata->inWarps, sizeof(TWWarp)*sdata->numInWarps);
      tempArray[sdata->numInWarps]=newSect;
      sdata->numInWarps++;
      delete[] sdata->inWarps;
      sdata->inWarps=tempArray;
      }
   return true;
}

Bool TWSector::addOutWarp(const TWWarp& newSect)
{
   if (sdata->numOutWarps==0)
      {
      sdata->outWarps = new TWWarp[1];
      if (sdata->outWarps==NULL)
         return false;
      sdata->numOutWarps++;
      sdata->outWarps[0]=newSect;
      }
   else
      {
      TWWarp *tempArray = new TWWarp[sdata->numOutWarps+1];
      if (tempArray==NULL)
         return false;
      memcpy(tempArray, sdata->outWarps, sizeof(TWWarp)*sdata->numOutWarps);
      tempArray[sdata->numOutWarps]=newSect;
      sdata->numOutWarps++;
      delete[] sdata->outWarps;
      sdata->outWarps=tempArray;
      }
   return true;
}

Bool TWSector::isExplored(void) const
{
   return sdata->explored;
}

void TWSector::setExplored(Bool expState)
{
   sdata->explored=expState;
}

Bool TWSector::hasInWarp(USHORT testSect) const
{
   if (sdata->numInWarps==0)
      return false;
   for (USHORT i=0; i<sdata->numInWarps; i++)
      {
      if (sdata->inWarps[i].sector()==testSect)
         return true;
      }
   return false;
}

Bool TWSector::hasOutWarp(USHORT testSect) const
{
   if (sdata->numOutWarps==0)
      return false;
   for (USHORT i=0; i<sdata->numOutWarps; i++)
      {
      if (sdata->outWarps[i].sector()==testSect)
         return true;
      }
   return false;
}

TWSectorArray::TWSectorArray(USHORT ns)
{
   numSects=ns;
   sects = new TWSector[numSects];
   if (sects==NULL)
      OutOfMemory();
}

TWSectorArray::TWSectorArray()
{
   numSects=0;
   sects=0;
}

TWSectorArray::~TWSectorArray()
{
   if (numSects>0)
      delete[] sects;
}

Bool TWSectorArray::setNumberOfSectors(USHORT ns)
{
   if (numSects>0)
      delete[] sects;
   numSects=ns;
   sects = new TWSector[numSects];
   if (sects==NULL)
      return false;
   else
      return true;
}

USHORT TWSectorArray::numberOfSectors(void) const
{
   return numSects;
}

TWSector& TWSectorArray::operator[](USHORT index)
{
   return sects[index];
}

TWSector& TWSectorArray::sectorAtPosition(USHORT index)
{
   return sects[index];
}

void TWSectorArray::readWarpData(ifstream& sectFile)
{
   char sectLine[256], tokens[7][10], *toktemp;
   USHORT i;
   Bool noMoreTokens;
   USHORT lineNum=0, numTokens, sectNum, currWarpSect;
   TWWarp currWarp;

   sectFile.seekg(0);
   while (!sectFile.eof())
      {
      sectFile.getline(sectLine, 255);
      if (sectFile.eof())
         break;
      lineNum++;
      noMoreTokens=false;
      numTokens=0;
      for (i=0; i<7; i++)
         {
         memset(tokens[i], 0, sizeof(tokens[i]));
         }
      toktemp=strtok(sectLine, " ");
      if (toktemp==NULL)
         continue;
      strcpy(tokens[numTokens], toktemp);
      numTokens++;
      while (!noMoreTokens)
         {
         toktemp=strtok(NULL, " ");
         if (toktemp==NULL)
            noMoreTokens=true;
         else
            {
            strcpy(tokens[numTokens], toktemp);
            numTokens++;
            }
         }
      sectNum=(USHORT)atoi(tokens[0]);
      if (sectNum==0)
         continue;
      if (sectNum>Globals::totalSectors)
         BadSector(tokens[0], lineNum, BADSECT_TOO_LARGE);
      if (sects[sectNum-1].isExplored())
         BadSector(tokens[0], lineNum, BADSECT_DUP_DATA);

      cout << "\b\b\b\b\b" << setw(5) << sectNum << flush;
      sects[sectNum-1].setExplored(true);
      for (i=1; i<numTokens; i++)
         {
         currWarpSect=(USHORT)atoi(tokens[i]);
         if (currWarpSect==0)
            BadSector(tokens[i], lineNum, BADSECT_NOT_INTEGER);
         if (currWarpSect>Globals::totalSectors)
            BadSector(tokens[i], lineNum, BADSECT_TOO_LARGE);
         currWarp.setSector(currWarpSect);
         currWarp.setOneWay(false);
         if (!sects[sectNum-1].addOutWarp(currWarp))
            OutOfMemory();
         currWarp.setSector(sectNum);
         if (!sects[currWarpSect-1].addInWarp(currWarp))
            OutOfMemory();
         }
      }
}

void TWSectorArray::setOneWayStatus(void)
{
   UCHAR warps;
   USHORT i, j, sectNum;
   TWWarp currWarp;

   for (i=0; i<Globals::totalSectors; i++)
      {
      cout << "\b\b\b\b\b" << setw(5) << (i+1) << flush;
      warps=sects[i].numberOfInWarps();
      for (j=0; j<warps; j++)
         {
         currWarp=sects[i].inWarpAtPosition(j);
         sectNum=currWarp.sector();
         if (sects[i].hasOutWarp(sectNum))
            sects[i].inWarpAtPosition(j).setOneWay(false);
         else
            sects[i].inWarpAtPosition(j).setOneWay(true);
         }
      warps=sects[i].numberOfOutWarps();
      for (j=0; j<warps; j++)
         {
         currWarp=sects[i].outWarpAtPosition(j);
         sectNum=currWarp.sector();
         if (sects[i].hasInWarp(sectNum))
            sects[i].outWarpAtPosition(j).setOneWay(false);
         else
            sects[i].outWarpAtPosition(j).setOneWay(true);
         }
      }
}

void TWSectorArray::findBubbles(TWBubbleSet &bubbleSet)
{
   for (i=0; i<Globals::totalSectors; i++)
      {
      #if (defined(__OS2__) || defined(__EMX__))
      if (Globals::pclass==3)
         {
         if (i%100==0)
            DosSleep(1);
         }
      #endif
      while (_kbhit())
         {
         if (_getch()==32)
            {
            fflush(stdin);
            return;
            }
         }
      cout << "\b\b\b\b\b" << setw(5) << (i+1) << flush;
      warps=sects[i].numberOfOutWarps();
      entrance=i+1;
      if (Globals::tunnelEntrance)
         {
         for (uniIterator=0; uniIterator<warps; uniIterator++)
            {
            currBubble = new TWBubble(entrance);
            if (currBubble==NULL)
               OutOfMemory();
            if (!currBubble->addSector(sects[i].outWarpAtPosition(uniIterator).sector()))
               OutOfMemory();
            bubbleCursor=1;
            uniSects = new USHORT[warps];
            if (uniSects==NULL)
               OutOfMemory();
            numUniSects=0;
            for (j=0; j<warps; j++)
               {
               if (j!=uniIterator)
                  {
                  sectNum=sects[i].outWarpAtPosition(j).sector();
                  uniSects[numUniSects]=sectNum;
                  numUniSects++;
                  }
               }
            traverseBubble(bubbleSet);
            }
         }
      else
         {
         uniMasks=(UCHAR)ipow(2L, (ULONG)warps);
         for (uniIterator=1; uniIterator<(uniMasks-1); uniIterator++)
            {
            currBubble = new TWBubble(entrance);
            if (currBubble==NULL)
               OutOfMemory();
            bubbleCursor=1;
            uniSects = new USHORT[warps];
            if (uniSects==NULL)
               OutOfMemory();
            numUniSects=0;
            for (j=0; j<warps; j++)
               {
               sectNum=sects[i].outWarpAtPosition(j).sector();
               if ((uniIterator<<(31-j)) >> 31)
                  {
                  uniSects[numUniSects]=sectNum;
                  numUniSects++;
                  }
               else
                  {
                  if (!currBubble->addSector(sectNum))
                     OutOfMemory();
                  }
               }
            traverseBubble(bubbleSet);
            }
         }
      }
}

Bool TWSectorArray::addWarps(USHORT previous, USHORT current)
{
   if ((currBubble->numberOfSectors()-1)>Globals::maxSectors)
      {
      return false;
      }
   UCHAR numWarps, i, j;
   TWWarp currWarp;
   USHORT sectNum;
   numWarps=sects[current-1].numberOfOutWarps();
   for (i=0; i<numWarps; i++)
      {
      currWarp=sects[current-1].outWarpAtPosition(i);
      sectNum=currWarp.sector();
      if (!sects[sectNum-1].isExplored())
         return false;
      if (currWarp.isOneWay())
         {
         if (!currBubble->addBackDoor(current))
            OutOfMemory();
         continue;
         }
      if (sectNum==previous)
         continue;
      if (currBubble->hasSector(sectNum))
         continue;
      for (j=0; j<numUniSects; j++)
         {
         if (sectNum==uniSects[j])
            return false;
         }
      if (!currBubble->addSector(sectNum))
         OutOfMemory();
      }
   numWarps=sects[current-1].numberOfInWarps();
   for (i=0; i<numWarps; i++)
      {
      currWarp=sects[current-1].inWarpAtPosition(i);
      sectNum=currWarp.sector();
      if (sectNum==previous)
         continue;
      if (currBubble->hasSector(sectNum))
         continue;
      for (j=0; j<numUniSects; j++)
         {
         if (sectNum==uniSects[j])
            return false;
         }
      if (!currBubble->addSector(sectNum))
         OutOfMemory();
      }
   return true;
}

void TWSectorArray::traverseBubble(TWBubbleSet& bubbleSet)
{
   previous=entrance;
   bubbleGood=true;
   bubbleDone=false;
   while (!bubbleDone)
      {
      sectNum=currBubble->sectorAtPosition(bubbleCursor);
      if (sectNum==0)
         break;
      else if (!sects[sectNum-1].isExplored())
         {
         bubbleGood=false;
         bubbleDone=true;
         }
      else if (!addWarps(previous, sectNum))
         {
         bubbleDone=true;
         bubbleGood=false;
         }
      else
         {
         if (Globals::maxEntrances>1)
            {
            if ((bubbleCursor+1>(USHORT)(currBubble->numberOfSectors() -
                (bubbleCursor))*Globals::entMultiplier) &&
                ((bubbleCursor+1)>Globals::minSectors) &&
                (currBubble->numberOfSectors()-(bubbleCursor+1) <
                Globals::maxEntrances))
               {
               tempBubble = new TWBubble(*currBubble);
               if (tempBubble==NULL)
                  OutOfMemory();
               tempCursor=bubbleCursor;
               while ((tempCursor+1)<tempBubble->numberOfSectors())
                  {
                  tempCursor++;
                  tsect=tempBubble->sectorAtPosition(tempCursor);
                  if (!tempBubble->addEntrance(tsect))
                     OutOfMemory();
                  }
               bubbleSet.addBubble(*tempBubble);
               delete tempBubble;
               }
            }
         }
      previous=sectNum;
      bubbleCursor++;
      if ((bubbleCursor+1)>currBubble->numberOfSectors())
         bubbleDone=true;
      }
   if (bubbleGood && currBubble->numberOfSectors()>Globals::minSectors)
      {
      bubbleSet.addBubble(*currBubble);
      }
   delete currBubble;
   delete[] uniSects;
}
