#include "TWTypes.hpp"
#include "TWWarp.hpp"

class ifstream;
class TWBubbleSet;

typedef struct TWSectorData
{
   TWWarp *inWarps;
   TWWarp *outWarps;
   UCHAR numInWarps;
   UCHAR numOutWarps;
   Bool explored;
};

class TWSector
{
   public:

   TWSector();
   TWSector(TWSector& copy);
   TWSector& operator=(const TWSector& assign);
   ~TWSector();

   UCHAR numberOfInWarps(void) const;
   UCHAR numberOfOutWarps(void) const;
   TWWarp& inWarpAtPosition(UCHAR);
   TWWarp& outWarpAtPosition(UCHAR);
   Bool addInWarp(const TWWarp& addition);
   Bool addOutWarp(const TWWarp& addition);
   Bool isExplored(void) const;
   void setExplored(Bool);
   Bool hasInWarp(USHORT) const;
   Bool hasOutWarp(USHORT) const;

   private:

   TWSectorData *sdata;
};

class TWSectorArray
{
   public:

   TWSectorArray(USHORT numSects);
   TWSectorArray();
   ~TWSectorArray();

   TWSector& sectorAtPosition(USHORT index);
   Bool setNumberOfSectors(USHORT numSects);
   TWSector& operator[](USHORT index);
   USHORT numberOfSectors(void) const;
   void readWarpData(ifstream& dataStream);
   void setOneWayStatus(void);
   void findBubbles(TWBubbleSet &bubbleSet);
   void traverseBubble(TWBubbleSet &bubbleSet);

   private:

   Bool addWarps(USHORT previous, USHORT sectNum);
   TWSector *sects;
   USHORT numSects, *uniSects, i, j, entrance, previous, bubbleCursor,
          tempCursor, sectNum, tsect;
   UCHAR numUniSects, warps, uniMasks;
   TWBubble *currBubble, *tempBubble;
   ULONG uniIterator;
   Bool bubbleGood, bubbleDone;
};
